using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Data.OleDb;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace VfpClient {
    public partial class VfpDataReader : DbDataReader {
        private Dictionary<string, VfpType> _columnVfpTypes;
        private readonly DbDataReader _dbDataReader;
        private static readonly Regex _guidRegex = new Regex(@"^(\{){0,1}[0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}(\}){0,1}$", RegexOptions.Compiled);

        public override int Depth {
            get {
                try {
                    return _dbDataReader.Depth;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override int FieldCount {
            get {
                try {
                    return _dbDataReader.FieldCount;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override bool HasRows {
            get {
                try {
                    return _dbDataReader.HasRows;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override bool IsClosed {
            get {
                try {
                    return _dbDataReader.IsClosed;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override int RecordsAffected {
            get {
                try {
                    return _dbDataReader.RecordsAffected;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override object this[string name] {
            get {
                try {
                    return GetValue(GetOrdinal(name));
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override object this[int ordinal] {
            get {
                try {
                    return GetValue(ordinal);
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public override int VisibleFieldCount {
            get {
                try {
                    return _dbDataReader.VisibleFieldCount;
                }
                catch (OleDbException e) {
                    throw new VfpException(e.Message, e);
                }
            }
        }

        public VfpDataReader(DbDataReader dbDataReader) {
            ArgumentUtility.CheckNotNull("dbDataReader", dbDataReader);

            _dbDataReader = dbDataReader;
        }

        public override void Close() {
            try {
                _dbDataReader.Close();
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override bool GetBoolean(int ordinal) {
            try {
                return (bool)Convert.ChangeType(this[ordinal], typeof(bool));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override byte GetByte(int ordinal) {
            try {
                return (byte)Convert.ChangeType(this[ordinal], typeof(byte));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override long GetBytes(int ordinal, long dataOffset, byte[] buffer, int bufferOffset, int length) {
            try {
                return _dbDataReader.GetBytes(ordinal, dataOffset, buffer, bufferOffset, length);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override char GetChar(int ordinal) {
            try {
                return _dbDataReader.GetChar(ordinal);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override long GetChars(int ordinal, long dataOffset, char[] buffer, int bufferOffset, int length) {
            try {
                return _dbDataReader.GetChars(ordinal, dataOffset, buffer, bufferOffset, length);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override string GetDataTypeName(int ordinal) {
            try {
                return _dbDataReader.GetDataTypeName(ordinal).ToVfpTypeName();
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override DateTime GetDateTime(int ordinal) {
            try {
                return _dbDataReader.GetDateTime(ordinal);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override decimal GetDecimal(int ordinal) {
            try {
                return (decimal)Convert.ChangeType(this[ordinal], typeof(decimal));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override double GetDouble(int ordinal) {
            try {
                return (double)Convert.ChangeType(this[ordinal], typeof(double));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override IEnumerator GetEnumerator() {
            return _dbDataReader.GetEnumerator();
        }

        public override Type GetFieldType(int ordinal) {
            try {
                return _dbDataReader.GetFieldType(ordinal);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override float GetFloat(int ordinal) {
            try {
                return (float)Convert.ChangeType(this[ordinal], typeof(float));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override Guid GetGuid(int ordinal) {
            try {
                return new Guid(GetString(ordinal));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override short GetInt16(int ordinal) {
            try {
                return (short)Convert.ChangeType(this[ordinal], typeof(short));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override int GetInt32(int ordinal) {
            try {
                return (int)Convert.ChangeType(this[ordinal], typeof(int));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override long GetInt64(int ordinal) {
            try {
                return (long)Convert.ChangeType(this[ordinal], typeof(long));
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override string GetName(int ordinal) {
            try {
                return _dbDataReader.GetName(ordinal).ToProperIfAllLowerCase();
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override int GetOrdinal(string name) {
            try {
                return _dbDataReader.GetOrdinal(name);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override DataTable GetSchemaTable() {
            try {
                var schema = _dbDataReader.GetSchemaTable();

                if (!schema.Columns.Contains("VfpType")) {
                    new SchemaFixer(this).Fix(schema);
                }

                return schema;
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override string GetString(int ordinal) {
            try {
                var value = _dbDataReader[ordinal];
                string stringValue;

                if (value != null && value.GetType() == typeof(byte[])) {
                    stringValue = Encoding.Default.GetString((byte[])value);
                }
                else {
                    stringValue = (string)value;
                }

                return !string.IsNullOrEmpty(stringValue) ? stringValue.TrimEnd() : stringValue;
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override object GetValue(int ordinal) {
            try {
                var value = _dbDataReader.GetValue(ordinal);

                if (value is string || (value is byte[] && IsStringType(ordinal))) {
                    var stringValue = GetString(ordinal);

                    if (!string.IsNullOrEmpty(stringValue) && _guidRegex.IsMatch(stringValue)) {
                        return new Guid(stringValue);
                    }

                    return stringValue;
                }

                return _dbDataReader.GetValue(ordinal);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override int GetValues(object[] values) {
            ArgumentUtility.CheckNotNull("values", values);

            try {
                var total = Math.Min(values.Length, VisibleFieldCount);

                for (var index = 0; index < total; index++) {
                    values[index] = IsDBNull(index) ? DBNull.Value : GetValue(index);
                }

                return total;
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        private bool IsStringType(int ordinal) {
            var columnName = GetName(ordinal).ToUpper();

            return GetColumnVfpTypes().ContainsKey(columnName) && GetColumnVfpTypes()[columnName].IsStringType();
        }

        private Dictionary<string, VfpType> GetColumnVfpTypes() {
            if (_columnVfpTypes == null) {
                _columnVfpTypes = GetSchemaTable().AsEnumerable()
                                                  .Select(x => new {
                                                      ColumnName = x.Field<string>("ColumnName").ToUpper(),
                                                      VfpType = (VfpType) x.Field<int>("VfpType")
                                                  })
                                                  .ToDictionary(x => x.ColumnName, x => x.VfpType);
            }

            return _columnVfpTypes;
        }

        public override bool IsDBNull(int ordinal) {
            try {
                return _dbDataReader.IsDBNull(ordinal);
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override bool NextResult() {
            try {
                return _dbDataReader.NextResult();
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        public override bool Read() {
            try {
                return _dbDataReader.Read();
            }
            catch (OleDbException e) {
                throw new VfpException(e.Message, e);
            }
        }

        protected override void Dispose(bool disposing) {
            if (disposing && _dbDataReader != null) {
                _dbDataReader.Dispose();
            }

            base.Dispose(disposing);
        }
    }
}